import sys
import powerfactory as pf
import numpy as np
import pandas as pd
app = pf.GetApplication()
if app is None:
    raise Exception("getting PowerFactory application failed")

#Simulation start/end time, combined with code on the end of the script
from datetime import datetime
# datetime object containing current date and time
now = datetime.now()
# dd/mm/YY H:M:S
dt_string_start = now.strftime("%d/%m/%Y %H:%M:%S")
#print("date and time =", dt_string)	


#defining function for setting up initial conditions and simulation
def setupSimultation(comInc, comSim):
    # Initializiation
    comInc.iopt_sim = "rms"
    comInc.iopt_show = 0
    comInc.iopt_adapt = 0
    comInc.start = 0
    # Defining simulation time
    comSim.tstop = 8

def runSimulation(comInc, comSim):
    app.EchoOff() #disables PF user interface
    comInc.Execute()
    app.EchoOn()
    comSim.Execute()
  
def addRecordedResult(elmRes, obj, param):
    if type(obj) is str:
        for elm in app.GetCalcRelevantObjects(obj):
            elmRes.AddVariable(elm, param)
    elif type(obj) is list:
        for elm in obj:
            elmRes.AddVariable(elm, param)
    else:
        elmRes.AddVariable(obj, param)

def clearSimEvents():
    faultFolder = app.GetFromStudyCase("Simulationsereignisse/Fehler.IntEvt")
    cont = faultFolder.GetContents()
    for obj in cont:
        obj.Delete()    


def addShcEvent(obj, sec, faultType, R, X):
    faultFolder = app.GetFromStudyCase("Simulationsereignisse/Fehler.IntEvt")
    event = faultFolder.CreateObject("EvtShc", obj.loc_name)
    event.p_target = obj
    event.time = sec
    event.i_shc = faultType
    if faultType == 0:
        event.R_f = R
        event.X_f = X

## returns the list of selected elements by name        
def getSelectedElements(All_elements,Wanted_elements): #takes a list of objects and a list of strings(names)
    elements = []
    for i in range(0,len(Wanted_elements)):
        for z in range(0,len(All_elements)):
            if str(All_elements[z].loc_name) == Wanted_elements[i]:
                elements.append(All_elements[z])
    if len(elements)==0:           
        raise Exception("Something went wrong, the returned list is empty")          
    return elements
    
    
app.ClearOutputWindow()

#--------------------Scripting cases---------------------------------------------
Sim_number = 0 #counter

output_size = 1320 # adjust to the total amount of planned simulations
Output_Iteration_number = np.zeros(output_size)
Output_Load_ratio = np.zeros(output_size)       
Output_Motor_A_share = np.zeros(output_size)
Output_Motor_B_share = np.zeros(output_size)
Output_Motor_C_share = np.zeros(output_size)
Output_Motor_D_share = np.zeros(output_size)
Output_Fault_bus = np.zeros(output_size)
Output_Instability = np.zeros(output_size)
Indexes_3s = np.zeros(output_size)
Motors_case_output = np.zeros(output_size)
Output_CVD = np.zeros(output_size)

Fault_buses_of_interest = ['4042','4043','4044','4041','4062','4031','4032','1041','1042','1043','1044','1045']
Fault_buses = []
All_buses = app.GetCalcRelevantObjects("*.ElmTerm")

Fault_buses = getSelectedElements(All_buses,Fault_buses_of_interest)


Load_names_of_interest = ['01','02','03','04',
                          '05','41','42','43','46','47','51']
loads = []   
All_loads = app.GetCalcRelevantObjects('*.ElmLod')
loads = getSelectedElements(All_loads,Load_names_of_interest)

DER_names_of_interest = ['DER(1)','DER(2)','DER(3)','DER(4)','DER(5)','DER(41)',
                         'DER(42)','DER(43)','DER(46)','DER(47)','DER(51)']
All_DERs = app.GetCalcRelevantObjects('*.ElmGenstat')   
DERs = []
DERs = getSelectedElements(All_DERs,DER_names_of_interest)

Buses_of_interest = ['1041','1042','1043','1045','4041',
                     '4042','4043','4046','4047','4051']
Buses = []
Buses = getSelectedElements(All_buses,Buses_of_interest)


Nordic = app.GetCalcRelevantObjects('Nordic.ElmNet')[0] ###CHANGED
All_folders = Nordic.GetContents()

Folders = []
Folders_of_interest = ['WECC CMPLDW (01)', 'WECC CMPLDW (02)', 
                       'WECC CMPLDW (03)', 'WECC CMPLDW (04)','WECC CMPLDW (05)', 
                       'WECC CMPLDW (41)', 'WECC CMPLDW (42)','WECC CMPLDW (43)',
                       'WECC CMPLDW (46)', 'WECC CMPLDW (47)','WECC CMPLDW (51)']
Folders = getSelectedElements(All_folders,Folders_of_interest)
          
Scripts = []
DSLs_motors = [] 

for i in range(0,len(Folders)):
    Scripts.append(Folders[i].GetContents('*.ComDpl',1)[0])

for i in range(0,len(Folders)):
    DSLs_motors.append(Folders[i].GetContents('Motor D dynamic model.ElmDsl',1)[0])


DSL_names_of_interest = ['DER (01)','DER (02)','DER (03)','DER (04)','DER (05)','DER (41)',
                         'DER (42)','DER (43)','DER (46)','DER (47)','DER (51)']

DSLs = []
All_DSLs = app.GetCalcRelevantObjects('*.ElmDsl')

for i in range(0,len(DSL_names_of_interest)):
    for z in range(0,len(All_DSLs)):
        if str(All_DSLs[z].chr_name) == DSL_names_of_interest[i]:
            DSLs.append(All_DSLs[z])



#Share of motors ABCD scenarios 1-10                #Case number/code
Motor_abcd_test = ( [0, 0, 0, 0],                   #0    'N'
                    [0.15, 0.15, 0.15, 0.15],       #1    '0'
                    [0.3, 0.1, 0.1, 0.1],           #2    'A'
                    [0.45, 0.05, 0.05, 0.05],       #3    'AA'
                    [0.1, 0.3, 0.1, 0.1],           #4    'B'
                    [0.05, 0.45, 0.05, 0.05],       #5    'BB'
                    [0.1, 0.1, 0.3, 0.1],           #6    'C'
                    [0.05, 0.05, 0.45, 0.05],       #7    'CC'
                    [0.1, 0.1, 0.1, 0.3],           #8    'D'
                    [0.05, 0.05, 0.05, 0.45],       #9    'DD'
                  )

Load_ratio_vector = [1, 0.95, 0.9, 0.85, 0.8, 0.75, 0.7, 0.65, 0.6, 0.55, 0.5]

#-------------Set up simulation cases----------------------------------------------------------
for x in range(0, 1): # 11
    Load_ratio = 0.6 #Load_ratio_vector[x]
    for y in range (0, 1): #10
        Motors_abcd = Motor_abcd_test[7]  
        #counter = 0
        for z in range (1, 12): #12
            Fault_bus = Fault_buses[z] 
            
            #-------------Edit load power-------------------------------------------------------

            
            P_nom = [600, 330, 260, 840, 720, 540, 400, 900, 700, 100, 800]
            Q_nom = [148, 71, 84, 252, 190, 131, 127, 254, 212, 44, 258]
            
            for i in range(0,len(loads)):  
                loads[i].plini = P_nom[i]*Load_ratio
                loads[i].qlini = Q_nom[i]*Load_ratio

            #---------------
            Sim_number = Sim_number + 1
            

            #--------------Edit DER power ---------------------------------------------------------- 
            
            
            for i in range(0,len(DERs)):
                DERs[i].sgn = 13.86
                DERs[i].pgini = 13.86
                DERs[i].Pmax_ucPU = 13.86
                DERs[i].Pmax = DERs[i].Pnom
                DERs[i].pgini = 13.86
                DERs[i].outserv = 1
         
        
            #--------------------Edit CMPLDW parameters----------------------------------------------    


            Motor_Elec_share = 0.15
            #Static_load = 1-sum(motors)

            V_stall = 0.7 # Load at which Motors D stall. V_stall_nom = 0.55
            F_rst = 0 # Fraction of Motor D that restarts after stall. F_rst_nom = 0.2          
            
            for i in range(0,len(DSLs_motors)):
                
                DSLs_motors[i].Vstall = V_stall
                DSLs_motors[i].Frst = F_rst
                DSLs_motors[i].Tstall = 0.033 #0.033
                #DSLs_motors[i].Xstall = 0.1 # def = 0.1
            
            
            
            for i in range(0,len(Scripts)):
                
                Scripts[i].SetInputParameterDouble('Pset', P_nom[i] - loads[i].plini + 0.1)
                Scripts[i].SetInputParameterDouble('Qset', Q_nom[i] - loads[i].qlini + 0.1)
                
                Scripts[i].SetInputParameterDouble('Fma', Motors_abcd[0])
                Scripts[i].SetInputParameterDouble('Fmb', Motors_abcd[1])
                Scripts[i].SetInputParameterDouble('Fmc', Motors_abcd[2])
                Scripts[i].SetInputParameterDouble('Fmd', Motors_abcd[3])
                Scripts[i].SetInputParameterDouble('Fel', 0.15)
                Scripts[i].Execute() 

            #-----------------------Faults and simulation setup---------------------------------------

            comInc = app.GetFromStudyCase("ComInc")
            comSim = app.GetFromStudyCase("ComSim")
            comInc.Execute() # Run initial conditions
            setupSimultation(comInc, comSim)

            clearSimEvents()
            faultFolder = app.GetFromStudyCase("Simulationsereignisse/Fehler")
            SC_element = Fault_bus # Current object taken from loop z

            time = 1 # 
            clearTime = 1.1
            faultType = 0  # Fault type 0 is 3-phase SC
            faulClear = 4  # Fault type 4 is fault clearing
            R = 20# # Fault resistance
            X = 0 # Fault reactance
            
            addShcEvent(SC_element, time, faultType, R, X)  # Add SC event
            addShcEvent(SC_element, clearTime, faulClear, R, X) # Add SC clearing

            runSimulation(comInc, comSim) # Run simulation

            #--------------------------Plotting------------------------------------------------




            setDesktop=app.GetGraphicsBoard()
            viPage = setDesktop.GetPage('Central_Voltages',1)
            oVi = viPage.GetVI('RMS Voltages','VisPlot',1)
            
            for i in range(0,len(Buses_of_interest)):
                oVi.AddVars(Buses[i], "m:u1")

            viPage.DoAutoScaleX()
            viPage.DoAutoScaleY()

            #-----------------------Save output window-------------------------------------------------

            graphPage=app.GetGraphicsBoard()
            
            for i in range(0,len(graphPage.GetContents('*.SetVipage'))):
                           if graphPage.GetContents('*.SetVipage')[i].loc_name == 'Central_Voltages':
                               VIGraph = graphPage.GetContents('*.SetVipage')[i]
 
            graphPage.Show(VIGraph)
            file_name=(r'C:\Users\aboricic\PLOTS' + '\\' + str(Sim_number) + 'RMS Voltages') # +graph.loc_name if 'for' is used
            graphPage.WriteWMF(file_name)

           ##---------------------Save CSV raw data-------------------------------------------------

            COMRES = app.GetFromStudyCase("ASCII Results Export.ComRes")
            app.PrintInfo(COMRES)
            # Adding a variable to COMRES and ElmRes so it gets exported to CSV
            ElmRes = app.GetCalcRelevantObjects("*.ElmRes")[0]
            
            
            for i in range(0,len(Buses)):        
                addRecordedResult(ElmRes, Buses[i], "m:u1")        
            
            COMRES.element = [] # reset values
            COMRES.variable = []
            
            COMRES.element = [ElmRes] + Buses
            COMRES.variable = ["b:tnow"] + len(Buses)*['m:u1']
                               

            COMRES.iopt_exp = 6
            COMRES.f_name = (r'C:\Users\aboricic\PLOTS' + '\\' + str(Sim_number) + 'TEST.csv')
            COMRES.Execute()

            #---------------------Analyze stability of current data/plots----------------------------
            
           
            voltages = pd.DataFrame(pd.read_csv(COMRES.f_name, skiprows=0, decimal='.'))
            
            DataSet = voltages.loc[2:].apply(pd.to_numeric)
            DataSet.columns = ['Time'] + Buses_of_interest
            DataSet.reset_index(drop=True,inplace=True)

            t_round = np.floor(DataSet['Time'])
            for i, time in enumerate(t_round):
                if time == 3:
                    index_3s = i
                    break

            app.PrintInfo(DataSet.iloc[index_3s:,1:])
              
            Instability = 0    
            if ((DataSet.iloc[index_3s:,1:] > 1.2).values.any() == True) or ((DataSet.iloc[index_3s:,1:] < 0.8).values.any() == True):
                Instability = 1
                             
            app.PrintInfo(Instability)    
                
            Motors_case = y    

            #--------------------Create overview excel file------------------------------------------
           
            Output_Iteration_number[Sim_number-1] = Sim_number            
            Output_Load_ratio[Sim_number-1] = Load_ratio         
            Output_Motor_A_share[Sim_number-1] = Motors_abcd[0]
            Output_Motor_B_share[Sim_number-1] = Motors_abcd[1]
            Output_Motor_C_share[Sim_number-1] = Motors_abcd[2]
            Output_Motor_D_share[Sim_number-1] = Motors_abcd[3]
            Output_Fault_bus[Sim_number-1] = int(Fault_bus.loc_name)
            Output_Instability[Sim_number-1] = Instability
            Indexes_3s[Sim_number-1] = index_3s
            Motors_case_output[Sim_number-1] = Motors_case
                
                 
            Data = {'Simulation Number': Output_Iteration_number,
                    'Load Ratio': Output_Load_ratio,
                    'Motor A share': Output_Motor_A_share,
                    'Motor B share': Output_Motor_B_share,
                    'Motor C share': Output_Motor_C_share,
                    'Motor D share': Output_Motor_D_share,
                    'Fault Bus': Output_Fault_bus,
                    'Instability': Output_Instability,
                    'Index_3sec': Indexes_3s,
                    'Motors_case': Motors_case_output}


            #-------------------Print in PF window-------------------------------------------------


            app.PrintInfo('Iteration ' + str(Sim_number) + ' completed successfully.')
            #app.PrintInfo('Load_ratio = ' + str(Load_ratio))
            #app.PrintInfo(SC_element)
           
            
#------------------------------Export simulation overview--------------------------------------------
DataFrame = pd.DataFrame(Data, columns = ['Simulation Number', 'Load Ratio', 'Motor A share', 'Motor B share', 'Motor C share', 
                                                      'Motor D share', 'Fault Bus','Instability','Index_3sec', 'Motors_case'])

DataFrame.to_excel (r'C:\Users\aboricic\PLOTS\Simulations Overview.xlsx', index = False, header=True)    

#------------------------------PowerFactory print in output window----------------------------------

app.PrintInfo('Script completed successfully!')

# Print start/end time of simulation
now = datetime.now()
dt_string_end = now.strftime("%d/%m/%Y %H:%M:%S")
app.PrintInfo('Start time:')
app.PrintInfo(dt_string_start) #from the start of the code
app.PrintInfo('End time:')
app.PrintInfo(dt_string_end)


